from model import Model
from record import Record
import config

__author__ = 'Emmanuel Turlay'
__copyright__ = 'Copyright 2017, Instacart'
__credits__ = ['Emmanuel Turlay', 'Mathieu Ripert']
__license__ = 'MIT'
__version__ = '0.5.2'
__maintainer__ = 'Emmanuel Turlay'
__email__ = 'emmanuel@instacart.com'
__status__ = 'Prototype'